@@ -32,7 +32,6 @@ import hashlib
 import errno
 import tempfile
 import functools
-import operator
 import re
 import contextlib
 import io
@@ -49,6 +48,8 @@ with contextlib.suppress(ImportError):
     import grp
 
 from . import matchers
+from . import masks
+from . import classes
 from .py37compat import best_realpath, lru_cache
 
 
@@ -57,44 +58,16 @@ __all__ = ['Path', 'TempDir']
 
 LINESEPS = ['\r\n', '\r', '\n']
 U_LINESEPS = LINESEPS + ['\u0085', '\u2028', '\u2029']
-U_NEWLINE = re.compile('|'.join(U_LINESEPS))
 B_NEWLINE = re.compile('|'.join(LINESEPS).encode())
-B_NL_END = re.compile(f'(?:{B_NEWLINE.pattern.decode()})$'.encode())
-U_NL_END = re.compile(f'(?:{U_NEWLINE.pattern})$')
+U_NEWLINE = re.compile('|'.join(U_LINESEPS))
+B_NL_END = re.compile(B_NEWLINE.pattern + b'$')
+U_NL_END = re.compile(U_NEWLINE.pattern + '$')
 
 
 class TreeWalkWarning(Warning):
     pass
 
 
-# from jaraco.functools
-def compose(*funcs):
-    compose_two = lambda f1, f2: lambda *args, **kwargs: f1(f2(*args, **kwargs))  # noqa
-    return functools.reduce(compose_two, funcs)
-
-
-class ClassProperty(property):
-    def __get__(self, cls, owner):
-        return self.fget.__get__(None, owner)()
-
-
-class multimethod:
-    """
-    Acts like a classmethod when invoked from the class and like an
-    instancemethod when invoked from the instance.
-    """
-
-    def __init__(self, func):
-        self.func = func
-
-    def __get__(self, instance, owner):
-        return (
-            functools.partial(self.func, owner)
-            if instance is None
-            else functools.partial(self.func, owner, instance)
-        )
-
-
 class Traversal:
     """
     Wrap a walk result to customize the traversal.
@@ -167,7 +140,7 @@ class Path(str):
         ns = {'module': module}
         return type(subclass_name, bases, ns)
 
-    @ClassProperty
+    @classes.ClassProperty
     @classmethod
     def _next_class(cls):
         """
@@ -390,7 +363,7 @@ class Path(str):
         """
         return self.splitext()[0]
 
-    @multimethod
+    @classes.multimethod
     def joinpath(cls, first, *others):
         """
         Join first to zero or more :class:`Path` components,
@@ -400,9 +373,7 @@ class Path(str):
 
         .. seealso:: :func:`os.path.join`
         """
-        if not isinstance(first, cls):
-            first = cls(first)
-        return first._next_class(first.module.join(first, *others))
+        return cls._next_class(cls.module.join(first, *others))
 
     def splitall(self):
         r"""Return a list of the path components in this path.
@@ -667,11 +638,7 @@ class Path(str):
         Default behavior is to overwrite any existing file.
         Call ``p.write_bytes(bytes, append=True)`` to append instead.
         """
-        if append:
-            mode = 'ab'
-        else:
-            mode = 'wb'
-        with self.open(mode) as f:
+        with self.open('ab' if append else 'wb') as f:
             f.write(bytes)
 
     def read_text(self, encoding=None, errors=None):
@@ -711,24 +678,22 @@ class Path(str):
 
         Parameters:
 
-          `text` - str/unicode - The text to be written.
+          `text` - str/bytes - The text to be written.
 
-          `encoding` - str - The Unicode encoding that will be used.
-              This is ignored if `text` isn't a Unicode string.
+          `encoding` - str - The text encoding used.
 
           `errors` - str - How to handle Unicode encoding errors.
               Default is ``'strict'``.  See ``help(unicode.encode)`` for the
-              options.  This is ignored if `text` isn't a Unicode
-              string.
+              options.  Ignored if `text` isn't a Unicode string.
 
           `linesep` - keyword argument - str/unicode - The sequence of
               characters to be used to mark end-of-line.  The default is
-              :data:`os.linesep`.  You can also specify ``None`` to
-              leave all newlines as they are in `text`.
+              :data:`os.linesep`.  Specify ``None`` to
+              use newlines unmodified.
 
           `append` - keyword argument - bool - Specifies what to do if
               the file already exists (``True``: append to the end of it;
-              ``False``: overwrite it.)  The default is ``False``.
+              ``False``: overwrite it).  The default is ``False``.
 
 
         --- Newline handling.
@@ -738,18 +703,13 @@ class Path(str):
         end-of-line sequence (see :data:`os.linesep`; on Windows, for example,
         the end-of-line marker is ``'\r\n'``).
 
-        If you don't like your platform's default, you can override it
-        using the `linesep=` keyword argument.  If you specifically want
-        ``write_text()`` to preserve the newlines as-is, use ``linesep=None``.
-
-        This applies to Unicode text the same as to 8-bit text, except
-        there are three additional standard Unicode end-of-line sequences:
-        ``u'\x85'``, ``u'\r\x85'``, and ``u'\u2028'``.
-
-        (This is slightly different from when you open a file for
-        writing with ``fopen(filename, "w")`` in C or ``open(filename, 'w')``
-        in Python.)
+        To override the platform's default, pass the `linesep=`
+        keyword argument. To preserve the newlines as-is, pass
+        ``linesep=None``.
 
+        This handling applies to Unicode text and bytes, except
+        with Unicode, additional non-ASCII newlines are recognized:
+        ``\x85``, ``\r\x85``, and ``\u2028``.
 
         --- Unicode
 
@@ -761,31 +721,37 @@ class Path(str):
         specified `encoding` (or the default encoding if `encoding`
         isn't specified).  The `errors` argument applies only to this
         conversion.
-
         """
         if isinstance(text, str):
             if linesep is not None:
                 text = U_NEWLINE.sub(linesep, text)
-            text = text.encode(encoding or sys.getdefaultencoding(), errors)
+            bytes = text.encode(encoding or sys.getdefaultencoding(), errors)
         else:
+            warnings.warn(
+                "Writing bytes in write_text is deprecated",
+                DeprecationWarning,
+                stacklevel=1,
+            )
             assert encoding is None
-            text = B_NEWLINE.sub(linesep.encode(), text)
-        self.write_bytes(text, append=append)
+            if linesep is not None:
+                text = B_NEWLINE.sub(linesep.encode(), text)
+            bytes = text
+        self.write_bytes(bytes, append=append)
 
-    def lines(self, encoding=None, errors='strict', retain=True):
+    def lines(self, encoding=None, errors=None, retain=True):
         r"""Open this file, read all lines, return them in a list.
 
         Optional arguments:
             `encoding` - The Unicode encoding (or character set) of
-                the file.  The default is ``None``, meaning the content
-                of the file is read as 8-bit characters and returned
-                as a list of (non-Unicode) str objects.
-            `errors` - How to handle Unicode errors; see help(str.decode)
-                for the options.  Default is ``'strict'``.
-            `retain` - If ``True``, retain newline characters; but all newline
-                character combinations (``'\r'``, ``'\n'``, ``'\r\n'``) are
-                translated to ``'\n'``.  If ``False``, newline characters are
-                stripped off.  Default is ``True``.
+                the file.  The default is ``None``, meaning use
+                ``locale.getpreferredencoding()``.
+            `errors` - How to handle Unicode errors; see
+                `open <https://docs.python.org/3/library/functions.html#open>`_
+                for the options.  Default is ``None`` meaning "strict".
+            `retain` - If ``True`` (default), retain newline characters,
+                but translate all newline
+                characters to ``\n``.  If ``False``, newline characters are
+                omitted.
 
         .. seealso:: :meth:`text`
         """
@@ -1098,7 +1064,7 @@ class Path(str):
         .. seealso:: :func:`os.chmod`
         """
         if isinstance(mode, str):
-            mask = _multi_permission_mask(mode)
+            mask = masks.compound(mode)
             mode = mask(self.stat().st_mode)
         os.chmod(self, mode)
         return self
@@ -1380,8 +1346,6 @@ class Path(str):
 
         Thereafter, the file at `filename` will have line numbers in it.
         """
-        import io
-
         if set(mode).intersection('wa+'):
             raise ValueError('Only read-only file modes can be used')
 
@@ -1437,7 +1401,7 @@ class Path(str):
         finally:
             backup_fn.remove_p()
 
-    @ClassProperty
+    @classes.ClassProperty
     @classmethod
     def special(cls):
         """
@@ -1569,7 +1533,7 @@ class Multi:
     def __iter__(self):
         return iter(map(self._next_class, self.split(os.pathsep)))
 
-    @ClassProperty
+    @classes.ClassProperty
     @classmethod
     def _next_class(cls):
         """
@@ -1598,7 +1562,7 @@ class TempDir(Path):
     .. seealso:: :func:`tempfile.mkdtemp`
     """
 
-    @ClassProperty
+    @classes.ClassProperty
     @classmethod
     def _next_class(cls):
         return Path
@@ -1621,82 +1585,6 @@ class TempDir(Path):
         self.rmtree()
 
 
-def _multi_permission_mask(mode):
-    """
-    Support multiple, comma-separated Unix chmod symbolic modes.
-
-    >>> oct(_multi_permission_mask('a=r,u+w')(0))
-    '0o644'
-    """
-    return compose(*map(_permission_mask, reversed(mode.split(','))))
-
-
-def _permission_mask(mode):
-    """
-    Convert a Unix chmod symbolic mode like ``'ugo+rwx'`` to a function
-    suitable for applying to a mask to affect that change.
-
-    >>> mask = _permission_mask('ugo+rwx')
-    >>> mask(0o554) == 0o777
-    True
-
-    >>> _permission_mask('go-x')(0o777) == 0o766
-    True
-
-    >>> _permission_mask('o-x')(0o445) == 0o444
-    True
-
-    >>> _permission_mask('a+x')(0) == 0o111
-    True
-
-    >>> _permission_mask('a=rw')(0o057) == 0o666
-    True
-
-    >>> _permission_mask('u=x')(0o666) == 0o166
-    True
-
-    >>> _permission_mask('g=')(0o157) == 0o107
-    True
-
-    >>> _permission_mask('gobbledeegook')
-    Traceback (most recent call last):
-    ValueError: ('Unrecognized symbolic mode', 'gobbledeegook')
-    """
-    # parse the symbolic mode
-    parsed = re.match('(?P<who>[ugoa]+)(?P<op>[-+=])(?P<what>[rwx]*)$', mode)
-    if not parsed:
-        raise ValueError("Unrecognized symbolic mode", mode)
-
-    # generate a mask representing the specified permission
-    spec_map = dict(r=4, w=2, x=1)
-    specs = (spec_map[perm] for perm in parsed.group('what'))
-    spec = functools.reduce(operator.or_, specs, 0)
-
-    # now apply spec to each subject in who
-    shift_map = dict(u=6, g=3, o=0)
-    who = parsed.group('who').replace('a', 'ugo')
-    masks = (spec << shift_map[subj] for subj in who)
-    mask = functools.reduce(operator.or_, masks)
-
-    op = parsed.group('op')
-
-    # if op is -, invert the mask
-    if op == '-':
-        mask ^= 0o777
-
-    # if op is =, retain extant values for unreferenced subjects
-    if op == '=':
-        masks = (0o7 << shift_map[subj] for subj in who)
-        retain = functools.reduce(operator.or_, masks) ^ 0o777
-
-    op_map = {
-        '+': operator.or_,
-        '-': operator.and_,
-        '=': lambda mask, target: target & retain ^ mask,
-    }
-    return functools.partial(op_map[op], mask)
-
-
 class Handlers:
     def strict(msg):
         raise
